# -----------------------------------------------------------------------------
#    HeartBeat - Yet Another HeartBeat Server
#    Copyright (C) 2011 Some Hackers In Town
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------------

import logging
from logging.handlers import RotatingFileHandler
from ConfigParser import NoSectionError


def setup_logger(config=None, logfile=None, debug=False):
    """
    Setup the Master channel of HeartBeat logger.
    """
    logparam = {}
    handlerparam = {}
    LOG_FORMAT = '[%(asctime)s] %(levelname)-8s : "%(message)s"'

    LOG_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
    LOG_MAX_BYTES = 2 * 1024 * 1024  # 2Mb
    LOG_BACKUP_COUNT = 5
    LOG_LEVEL = 1
    formatter = logging.Formatter(LOG_FORMAT)
    logparam['format'] = LOG_FORMAT
    logparam['level'] = int(LOG_LEVEL)
    logparam['datefmt'] = LOG_DATE_FORMAT
    handlerparam['maxBytes'] = LOG_MAX_BYTES
    handlerparam['backupCount'] = LOG_BACKUP_COUNT
    if config:
        try:
            logparam['level'] = int(config.get("Log", "level"))
            logparam['datefmt'] = config.get("Log", "date_format")
            handlerparam['maxBytes'] = int(config.get("Log", "max_size"))
            handlerparam['backupCount'] = int(config.get("Log", "backups"))
        except NoSectionError:
            print "Malform config file. 'Log' section is missing."
            print "Skipping."

    logging.basicConfig(**logparam)
    logger = logging.getLogger("HeartBeat")

    if logfile:
        LOG_FILENAME = logfile
        handler = RotatingFileHandler(
        LOG_FILENAME, **handlerparam)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    logger.propagate = debug
    return logger



# -----------------------------------------------------------------------------
#    HeartBeat - Yet Another HeartBeat Server
#    Copyright (C) 2011 Some Hackers In Town
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------------
import socket
import hashlib
import threading
import SocketServer

from hbeat.logging import logger


STATUS = {
    1: "OK",
    }


class HBProtocol (object):
    """
    HeartBeat protocol v1
    """
    status = 0
    host = None
    secret = None

    def __init__(self, data=None, status=0,
                 host=None, secret=None):
        if data:
            self.check(data)
            self.data = data
        self.host = host
        self.secret = secret
        self.status = status

    def check(self, data):
        """
        Check the integration of data.
        """
        if len(data) == 48:
            return True
        raise ValueError("'data' should have 48 byte of length.")

    def parse_data(self):
        """
        Return the status code for current data.
        """
        self.check(self.data)
        if self.data:
            self.status = ord(self.data[15])
            self.host = self.data[:15]
            self.hash_ = self.data[16:]
            return self.status, self.host.strip('\0'), self.hash_
        return None, None, None

    def create_packet(self):
        """
        Create a HeartBeat packet from status,
        host and secret key.
        """
        packet = None
        if self.host and self.status and self.secret:
            if len(self.host) <= 15:
                space = "\0" * (15 - len(self.host))
                if space:
                    packet = "%s%s" % (self.host, space)
                else:
                    packet = self.host

            else:
                raise TypeError("The size of 'host' should be less that" +
                                "15 charachter")
            if 1 <= self.status <= 255:
                packet = "%s%s" % (packet, chr(self.status))
            else:
                raise TypeError("'status' should be between 0 and 255.")

            m = hashlib.md5()
            m.update("%s%s" % (packet, self.secret))
            packet = "%s%s" % (packet, m.hexdigest())
            return packet
        raise ValueError("Missing some parameter to create packet.")

    def is_valid(self, secret):
        """
        check the current packet checksum.
        """
        secstring = "%s%s" % (self.data[:16],
                              secret)
        m = hashlib.md5()
        m.update(secstring)
        if self.data[16:] == m.hexdigest():
            return True
        else:
            return False


class TCPHandler(SocketServer.BaseRequestHandler, object):
    """
    The RequestHandler class for HearBeat server.

    It is instantiated once per connection to the server.
    """

    def __init__(self, *args, **kwargs):
        self.db = kwargs['db']
        del kwargs['db']
        super(TCPHandler, self).__init__(*args, **kwargs)

    def handle(self):
        self.data = self.request.recv(48).strip()
        packet = HBProtocol(data=self.data)
        status, host, hash_ = packet.parse_data()
        if not host.lower() in self.db.keys():
            self.request.send(chr(1))
            logger.warn("Invalid Host: '%s' IP: %s" % (host,
                                                     self.client_address[0]))
        else:

            if not packet.is_valid(self.db[host]["secret"]):
                logger.warn("Invalid packet form Host: '%s' IP: %s" % (host,
                                                    self.client_address[0]))
                self.request.send(chr(1))
            else:
                logger.info("Host: '%s' IP: %s Status: %s" % (host,
                                                        self.client_address[0],
                                                        STATUS[status]))
                self.request.send(chr(0))


class ThreadedTCPRequestHandler(SocketServer.BaseRequestHandler):

    def handle(self):
        data = self.request.recv(1024)
        cur_thread = threading.currentThread()
        response = "%s: %s" % (cur_thread.getName(), data)
        self.request.send(response)


class ThreadedTCPServer(SocketServer.ThreadingMixIn,
                        SocketServer.TCPServer):
    pass


class TCPServer (SocketServer.TCPServer, object):

    def __init__(self, address, handler, db):
        super(TCPServer, self).__init__(address, handler)
        self.db = db

    def finish_request(self, request, client_address):
        """Finish one request by instantiating RequestHandlerClass."""
        self.RequestHandlerClass(request, client_address,
                                 self, db=self.db)


class ThreadedHeartBeatServer(object):
    """
    Threaded HeartBeat Server class.
    """

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.server = ThreadedTCPServer((host, int(port)),
                                        ThreadedTCPRequestHandler)

    def serve_forever(self):
        self.server.serve_forever()

    def start(self):
        server_thread = threading.Thread(target=self.server.serve_forever)
        server_thread.setDaemon(True)
        server_thread.start()


class HeartBeatServer(object):
    """
    HeartBeat Server class.
    """

    def __init__(self, host, port, db):
        self.host = host
        self.port = port
        self.server = TCPServer((host, int(port)),
                                TCPHandler, db=db)

    def serve_forever(self):
        self.server.serve_forever()



# -----------------------------------------------------------------------------
#    HeartBeat - Yet Another HeartBeat Server
#    Copyright (C) 2011 Some Hackers In Town
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------------


import os
import sys
import stat
import signal
import atexit
from pwd import getpwnam
from ConfigParser import ConfigParser
from ConfigParser import NoSectionError


class HeartBeatDaemon (object):
    """
    Daemon class of HeartBeat, this class runs the HeartBeat in a daemon
    state (by default). You can think about this class as the main
    class of HeartBeat. This class use the *option* parameter for its
    configuration. *option* parameter is an object that created with
    optparser class and contains the command line parameters.

    if user decide to run the HeartBeat in background, then all of the
    IO transactions will redirect to given *stdin*, *stdout* and
    *stderr*.
    """

    def __init__(self, options, stdin='/dev/null', stdout='/dev/null',
                 stderr='/dev/null'):

        from hbeat.logging.base import setup_logger

        self.options = options

        # PID configurations ===================================
        self.piddir = options.piddir.rstrip("/")

        # debbox pid files
        self.mpid = "/".join((self.piddir, "heartbeat.pid"))

        # creating configuration object ========================
        #: confifgurewr
        self.config = ConfigParser()
        if os.path.exists(self.options.conf):
            self.config.read(self.options.conf)
        else:
            raise self.CantFindConfigFile()

        try:
            self.db = {}
            for line in file(self.options.db).readlines():
                if not line.strip(" ").startswith("#"):
                    name, secret = line.split(":")
                    self.db[name.strip().lower()] = {"secret": secret.strip(),
                                              }

        except ValueError:
            print "Error: Check the syntax of hosts.db file."
            sys.exit(0)

        except OSError:
            print "Error: Access problem on %s" % self.options.db
            sys.exit(0)

        # setting standard IO
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr

        # Webserver should run under which user and group
        try:
            self.slave_user = self.config.get("User", "user", "heartbeat")
        except NoSectionError:
            print "Error: Can't find a suitable username in config file."
            sys.exit(1)

        try:
            self.slave_group = self.config.get("User", "group", "heartbeat")
        except NoSectionError:
            print "Error: Can't find a suitable group name in config file."
            sys.exit(1)

        # Setting up log directory
        self.logfolder = self.config.get("Log", "folder")
        if not os.path.exists(self.logfolder):
            try:
                os.makedirs(self.logfolder)
            except OSError, e:
                print e
                sys.exit(1)

        try:
            uid = getpwnam(self.slave_user)[2]
            gid = getpwnam(self.slave_user)[3]
            os.chown(self.logfolder, uid, gid)
            os.chmod(self.logfolder,
                     stat.S_IXUSR | stat.S_IRUSR | stat.S_IWUSR | \
                     stat.S_IXGRP | stat.S_IWGRP | stat.S_IRGRP)

            setup_logger(self.config, "/".join((self.logfolder,
                                                "heartbeat.log")),
                                        self.options.debug)
        except OSError, e:
            print "%s" % e
            print "Note: Did you forget to run heartbeatd as root?"
            sys.exit(1)

        # we import logger here because it should import after
        # setup_logger called
        from hbeat.logging import logger

        self.logger = logger

        # Registering a cleanup method
        atexit.register(self.__cleanup__)

    def __cleanup__(self):
        """
        HeartBeat destructor.
        """
        # removing servers pid files on exit
        if os.path.exists(self.mpid):
            try:
                os.remove(self.mpid)
            except OSError, e:
                self.logger.info("Error msg: %s" % e)
                raise

    def _status(self):
        """
        checking for heartbeat processes.
        """
        if not os.path.exists(self.mpid):
            return False

        self._masterpid = file(self.mpid).readlines()[0]
        if os.path.exists("/proc/%s" % self._masterpid):
            return True

        else:
            try:
                os.remove(self.mpid)
            except OSError:
                pass
            return False

    def start(self):
        """
        Start the HeartBeat server. all the daemon forking process runs here.
        """
        from hbeat.server import HeartBeatServer

        if self._status():
            print "Debbox is already running."
            return

        # Daemonizing Process ==============================
        # Please read about how a daemon work before asking
        # questions

        if not self.options.foreground:
            pid = None
            try:
                pid = os.fork()
            except OSError:
                raise self.CantFork("Can't create the master process.")

            if pid > 0:
                # Exist from parent
                sys.exit(0)

            os.umask(027)
            try:
                self._sid = os.setsid()
            except OSError:
                raise

            self._masterpid = os.getpid()
            self.logger.info("Master process at %s" % self._masterpid)

            file(self.mpid, "w+").write(str(self._masterpid))

            uid = getpwnam(self.slave_user)[2]
            gid = getpwnam(self.slave_user)[3]

            uid = getpwnam(self.slave_user)[2]
            os.chown(self.mpid, uid, gid)
            os.setuid(int(uid))
            #os.setgid(int(gid))
            os.umask(027)

        if self.options.foreground:
            atexit.register(self.stop)

        server = HeartBeatServer(self.options.host,
                                 self.options.port,
                                 db=self.db)

        print "Starting HeartBeat server . . ."
        if not self.options.debug or not self.options.foreground:
                self.io_redirect()

        signal.signal(signal.SIGUSR1, self._usr1_handler)
        server.serve_forever()

    def stop(self):
        """
        Stop the HeartBeat server, and clean the environment with
        removing any temporary files and pid files.
        """
        if not self.options.foreground:
            if not self._status():
                self.logger.info("HeartBeat is not running. . .")
                print "HeatBeat is not running. . ."
                return

            mpid = file(self.mpid).readlines()[0]
            self.__cleanup__()
            print "Stopping master process."
            os.kill(int(mpid), 15)

        sys.exit(0)

    def status(self):
        """
        Return the status of HeartBeat server.
        """
        if os.path.exists(self.mpid):
            print "HeartBeat Master process is running with '%s' pid" % \
                  file(self.mpid).readlines()[0]
        else:
            print "Debbox Master is not running."

    def io_redirect(self):
        """
        Redirect all the IO to given IOs in the constructor. and skip the
        redirecting process if daemon run in foreground ar debbug mode.
        """
        if not self.options.debug or not self.options.foreground:
            # Redirecting standard I/O to nowhere
            sys.stdout.flush()
            sys.stderr.flush()
            si = file(self.stdin, 'r')
            so = file(self.stdout, 'a+')
            se = file(self.stderr, 'a+', 0)
            os.dup2(si.fileno(), sys.stdin.fileno())
            os.dup2(so.fileno(), sys.stdout.fileno())
            os.dup2(se.fileno(), sys.stderr.fileno())

    def _usr1_handler(self, signum, frame):
        """
        SIGUSR1 handler. HeartBeat will treat SIGUSR1 just like SIGTERM.
        """

        #os.waitpid(self._slavepid, 0)
        if self.options.foreground:
            mpid = os.getpid()
            os.kill(int(mpid), 15)
        self.stop()

    class CantFork (Exception):
        """
        This exception will raise if daemon can't for a new process.
        """
        pass

    class CantFindConfigFile (Exception):
        """
        This exception will raise if daemon can't find the Debbox
        main configuration file.
        """
        pass



#!/usr/bin/env python
from distutils.core import setup

setup(name='HeartBeat',
      version='0.1.0',
      description='Yet Another HeartBeat',
      author='Sameer Rahmani',
      author_email='lxsameer@gnu.org',
      url='http://git.lxsameer.com/?p=heartbeat.git;a=summary',
      license='GPL v2',
      scripts=["heartbeat", "heartbeatd"],
      data_files=[("/usr/share/man/man1/", ["man/heartbeatd.1"]),
                  ("/usr/share/man/man1/", ["man/heartbeat.1"]),
                  ("/etc/heartbeat/", ["conf/heartbeat.conf"]),
                  ("/etc/heartbeat/", ["conf/hosts.db"]),
                  ("/etc/init.d/", ["init/heartbeatd"]),
                  ],
      packages=['hbeat', 'hbeat.logging'],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Utilities',
          ]
)
